import asyncio
from typing import List
from reaktion.atoms.combination.base import CombinationAtom
from reaktion.events import EventType, OutEvent
import logging

logger = logging.getLogger(__name__)


class FilterAtom(CombinationAtom):
    # TODO: Rename to match as it maps to a specifc type
    complete: List[bool] = [False, False]

    async def run(self):
        try:
            while True:
                event = await self.get()

                if event.type == EventType.ERROR:
                    for index, stream in enumerate(self.node.outstream):
                        await self.transport.put(
                            OutEvent(
                                handle=f"return_{index}",
                                type=EventType.ERROR,
                                value=event.value,
                                source=self.node.id,
                                caused_by=[event.current_t],
                            )
                        )
                    break

                if event.type == EventType.NEXT:
                    value = event.value[0]
                    real_value = value["value"]
                    index = value["use"]
                    await self.transport.put(
                        OutEvent(
                            handle=f"return_{index}",
                            type=EventType.NEXT,
                            value=(real_value,),
                            source=self.node.id,
                            caused_by=[event.current_t],
                        )
                    )

                if event.type == EventType.COMPLETE:
                    for index, stream in enumerate(self.node.outstream):
                        await self.transport.put(
                            OutEvent(
                                handle=f"return_{index}",
                                type=EventType.COMPLETE,
                                source=self.node.id,
                                caused_by=[event.current_t],
                            )
                        )
                    break

        except asyncio.CancelledError as e:
            logger.warning(f"Atom {self.node} is getting cancelled")
            raise e

        except Exception as e:
            logger.exception(f"Atom {self.node} excepted")
            raise e



import logging
from typing import Callable, Dict
import asyncio
from pydantic import BaseModel, Field

from fluss.api.schema import (
    ArgNodeFragment,
    ArkitektNodeFragment,
    FlowFragment,
    LocalNodeFragment,
    ReactiveNodeFragment,
    ArkitektFilterNodeFragment,
    ReturnNodeFragment,
    arun,
    asnapshot,
    atrack,
    acondition_snapshot,
    astart_trace,
    atrace,
)
from reaktion.atoms.transport import AtomTransport

from reaktion.atoms.utils import atomify
from reaktion.contractors import NodeContractor, arkicontractor
from reaktion.events import EventType, InEvent, OutEvent

from reaktion.utils import connected_events
from rekuest.actors.base import Actor
from rekuest.api.schema import (
    AssignationStatus,
    ProvisionStatus,
    ReservationFragment,
    ReservationStatus,
    ProvisionMode,
    NodeFragment,
)
from rekuest.messages import Assignation, Provision
from rekuest.postmans.utils import RPCContract, ContractStatus
from typing import Any
from rekuest.collection.collector import AssignationCollector, ActorCollector
from rekuest.actors.transport.types import AssignTransport
from rekuest.actors.types import Assignment, Passport
from rekuest.actors.transport.local_transport import ProxyActorTransport

logger = logging.getLogger(__name__)


class NodeState(BaseModel):
    latestevent: OutEvent


class FlowActor(Actor):
    definition: NodeFragment
    is_generator: bool = False
    flow: FlowFragment
    agent: Any
    contracts: Dict[str, RPCContract] = Field(default_factory=dict)
    expand_inputs: bool = False
    shrink_outputs: bool = False
    provided = False
    is_generator: bool = False
    arkitekt_contractor: NodeContractor = arkicontractor
    snapshot_interval: int = 40
    condition_snapshot_interval: int = 40
    contract_states: Dict[str, ContractStatus] = Field(default_factory=dict)
    contract_t: int = 0

    # Functionality for running the flow

    # Assign Related Functionality
    run_mutation: Callable = arun
    snapshot_mutation: Callable = asnapshot
    track_mutation: Callable = atrack

    start_trace_mutation: Callable = astart_trace
    condition_snapshot_mutation: Callable = acondition_snapshot
    trace_mutation: Callable = atrace

    atomifier: Callable = atomify
    """ Atomifier is a function that takes a node and returns an atom """

    run_states: Dict[
        str,
        Dict[str, NodeState],
    ] = Field(default_factory=dict)

    reservation_state: Dict[str, ReservationFragment] = Field(default_factory=dict)
    _lock = None
    _condition = None

    async def on_provide(self, passport: Passport):
        self._lock = asyncio.Lock()

        self._condition = await self.start_trace_mutation(
            provision=passport.provision,
            flow=self.flow,
            snapshot_interval=self.condition_snapshot_interval,
        )

        [x for x in self.flow.graph.nodes if isinstance(x, ArgNodeFragment)][0]
        [x for x in self.flow.graph.nodes if isinstance(x, ReturnNodeFragment)][0]

        arkitektNodes = [
            x
            for x in self.flow.graph.nodes
            if isinstance(x, ArkitektNodeFragment)
            or isinstance(x, ArkitektFilterNodeFragment)
        ]

        arkitekt_contracts = {
            node.id: await self.arkitekt_contractor(node, self)
            for node in arkitektNodes
        }

        self.contracts = {**arkitekt_contracts}
        futures = [contract.aenter() for contract in self.contracts.values()]
        await asyncio.gather(*futures)

    async def on_local_log(self, reference, *args, **kwargs):
        logger.log(f"Contract log for {reference} {args} {kwargs}")

    async def on_local_change(
        self,
        reference,
        status: ProvisionStatus = None,
        message: str = None,
        mode: ProvisionMode = None,
    ):
        await self.trace_mutation(
            condition=self._condition, source=reference, value=str(status)
        )

    async def on_contract_change(
        self, state: ContractStatus = None, reference: str = None
    ):
        logger.debug(f"Contract change {reference}: {state} ")

        async with self._lock:
            if reference:
                self.contract_states[reference] = await self.trace_mutation(
                    condition=self._condition,
                    source=reference,
                    state=state,
                    value="Nothing",
                )

            inactive_contracts = [
                res
                for res in self.contracts.values()
                if res.state != ContractStatus.ACTIVE
            ]

            if len(inactive_contracts) > 0:
                await self.aset_status(
                    status=ProvisionStatus.CRITICAL,
                    message="One or more contracts are inactive",
                )
                logger.warning("Contract is inactive")

            else:
                await self.aset_status(
                    status=ProvisionStatus.ACTIVE,
                )

            if self.contract_t % self.snapshot_interval == 0:
                await self.condition_snapshot_mutation(
                    condition=self._condition,
                    events=list(self.contract_states.values()),
                )

            self.contract_t += 1

    async def on_assign(
        self,
        assignment: Assignment,
        collector: AssignationCollector,
        transport: AssignTransport,
    ):
        run = await self.run_mutation(
            assignation=assignment.assignation,
            flow=self.flow,
            snapshot_interval=self.snapshot_interval,
        )
        print(self.is_generator)

        await transport.log(level="INFO", message="Starting")

        t = 0
        state = {}
        await self.snapshot_mutation(run=run, events=list(state.values()), t=t)

        try:
            event_queue = asyncio.Queue()

            atomtransport = AtomTransport(queue=event_queue)

            argNode = [
                x for x in self.flow.graph.nodes if isinstance(x, ArgNodeFragment)
            ][0]
            returnNode = [
                x for x in self.flow.graph.nodes if isinstance(x, ReturnNodeFragment)
            ][0]

            participatingNodes = [
                x
                for x in self.flow.graph.nodes
                if isinstance(x, ArkitektNodeFragment)
                or isinstance(x, ArkitektFilterNodeFragment)
                or isinstance(x, ReactiveNodeFragment)
                or isinstance(x, LocalNodeFragment)
            ]

            await transport.log(level="INFO", message="Set up the graph")

            stream = self.flow.graph.args
            stream_keys = []
            for i in stream:
                stream_keys.append(i.key)

            global_keys = []
            for i in self.flow.graph.globals:
                global_keys.append(i.port.key)

            globalMap: Dict[str, Dict[str, Any]] = {}
            streamMap: Dict[str, Any] = {}

            assert len(self.definition.args) == len(
                assignment.args
            ), "Wrong number of args"

            for port, arg in zip(self.definition.args, assignment.args):
                if port.key in stream_keys:
                    streamMap[port.key] = arg
                if port.key in global_keys:
                    for i in self.flow.graph.globals:
                        if i.port.key == port.key:
                            for map in i.to_keys:
                                nodeid, key = map.split(".")
                                globalMap.setdefault(nodeid, {})[key] = arg

            async def ass_log(assignation: Assignation, level, message):
                await transport.log(level, message)
                logging.info(f"{assignation}, {message}")

            atoms = {
                x.id: self.atomifier(
                    x,
                    atomtransport,
                    self.contracts.get(x.id, None),
                    globalMap.get(x.id, {}),
                    assignment,
                    alog=ass_log,
                )
                for x in participatingNodes
            }

            await transport.log(level="INFO", message="Atomification complete")

            await asyncio.gather(*[atom.aenter() for atom in atoms.values()])
            tasks = [asyncio.create_task(atom.start()) for atom in atoms.values()]
            logger.info("Starting all Atoms")
            value = [streamMap[key] for key in stream_keys]

            initial_event = OutEvent(
                handle="return_0",
                type=EventType.NEXT,
                source=argNode.id,
                value=value,
                caused_by=[t],
            )
            initial_done_event = OutEvent(
                handle="return_0",
                type=EventType.COMPLETE,
                source=argNode.id,
                caused_by=[t],
            )

            logger.info(f"Putting initial event {initial_event}")

            await event_queue.put(initial_event)
            await event_queue.put(initial_done_event)

            edge_targets = [e.target for e in self.flow.graph.edges]
            nodes_without_instream = [
                x
                for x in participatingNodes
                if len(x.instream[0]) == 0 and x.id not in edge_targets
            ]

            for node in nodes_without_instream:
                assert node.id in atoms, "Atom not found. Should not happen."
                atom = atoms[node.id]

                initial_event = InEvent(
                    target=node.id,
                    handle="arg_0",
                    type=EventType.NEXT,
                    value=[],
                    current_t=t,
                )
                done_event = InEvent(
                    target=node.id,
                    handle="arg_0",
                    type=EventType.COMPLETE,
                    current_t=t,
                )

                await atom.put(initial_event)
                await atom.put(done_event)

            complete = False

            returns = []

            while not complete:
                event: OutEvent = await event_queue.get()
                event_queue.task_done()

                if self.flow.brittle:
                    if event.type == EventType.ERROR:
                        raise event.value

                track = await self.track_mutation(
                    run=run,
                    source=event.source,
                    handle=event.handle,
                    caused_by=event.caused_by,
                    value=event.value
                    if event.value and not isinstance(event.value, Exception)
                    else str(event.value),
                    type=event.type,
                    t=t,
                )
                state[event.source] = track.id

                # We tracked the events and proceed

                if t % self.snapshot_interval == 0:
                    await self.snapshot_mutation(
                        run=run, events=list(state.values()), t=t
                    )

                # Creat new events with the new timepoint
                spawned_events = connected_events(self.flow.graph, event, t)
                # Increment timepoint
                t += 1
                # needs to be the old one for now
                if not spawned_events:
                    logger.warning(f"No events spawned from {event}")

                for spawned_event in spawned_events:
                    logger.info(f"-> {spawned_event}")

                    if spawned_event.target == returnNode.id:
                        if spawned_event.type == EventType.NEXT:
                            returns = spawned_event.value
                            if self.is_generator:
                                print("Yielded")
                                await transport.change(
                                    status=AssignationStatus.YIELD,
                                    returns=returns,
                                )

                        if spawned_event.type == EventType.ERROR:
                            await self.snapshot_mutation(
                                run=run, events=list(state.values()), t=t
                            )
                            raise spawned_event.value

                        if spawned_event.type == EventType.COMPLETE:
                            await self.snapshot_mutation(
                                run=run, events=list(state.values()), t=t
                            )
                            complete = True
                            if not self.is_generator:
                                await transport.change(
                                    status=AssignationStatus.RETURNED,
                                    returns=returns,
                                )
                            else:
                                print("Done")
                                await transport.change(
                                    status=AssignationStatus.DONE,
                                )

                            logger.info("Done ! :)")

                    else:
                        assert (
                            spawned_event.target in atoms
                        ), "Unknown target. Your flow is connected wrong"
                        if spawned_event.target in atoms:
                            await atoms[spawned_event.target].put(spawned_event)

            for task in tasks:
                task.cancel()

            await asyncio.gather(*tasks, return_exceptions=True)
            logging.info("Collecting...")
            await self.collector.collect(assignment.id)
            logging.info("Done ! :)")

        except asyncio.CancelledError:
            for task in tasks:
                task.cancel()
            await self.snapshot_mutation(run=run, events=list(state.values()), t=t)

            try:
                await asyncio.wait_for(
                    asyncio.gather(*tasks, return_exceptions=True), timeout=4
                )
            except asyncio.TimeoutError:
                pass

            await self.collector.collect(assignment.id)
            await transport.change(status=AssignationStatus.CANCELLED)

        except Exception as e:
            logging.critical(f"Assignation {assignment} failed", exc_info=True)
            await self.snapshot_mutation(run=run, events=list(state.values()), t=t)
            await transport.log(message="Starting", level=AssignationStatus.ERROR)

            await self.collector.collect(assignment.id)
            await transport.change(
                status=AssignationStatus.CRITICAL,
                message=repr(e),
            )

    async def on_unprovide(self):
        for contract in self.contracts.values():
            await contract.aexit()



from typing import Protocol, runtime_checkable
from rekuest.postmans.utils import RPCContract, arkiuse, mockuse, actoruse
from fluss.api.schema import (
    ArkitektNodeFragment,
    FlowNodeFragmentBaseArkitektNode,
    LocalNodeFragment,
)
from rekuest.api.schema import afind, ReserveBindsInput, amytemplatefor, NodeScope
from rekuest.postmans.vars import get_current_postman
from rekuest.structures.registry import get_current_structure_registry
from rekuest.actors.base import Actor


@runtime_checkable
class NodeContractor(Protocol):
    async def __call__(self, node: ArkitektNodeFragment, actor: Actor) -> RPCContract:
        ...


async def arkicontractor(node: ArkitektNodeFragment, actor: Actor) -> RPCContract:
    try:
        template = await amytemplatefor(
            hash=node.hash, instance_id=actor.agent.instance_id
        )

        return actoruse(
            template=template,
            supervisor=actor,
            reference=node.id,
            state_hook=actor.on_contract_change,
            assign_timeout=node.assign_timeout or None,
            yield_timeout=node.yield_timeout or None,
            max_retries=node.max_retries,
            retry_delay_ms=node.retry_delay,
        )
    except Exception as e:
        arkinode = await afind(hash=node.hash)
        assert (
            arkinode.scope == NodeScope.GLOBAL
        ), "Non GlobalNodes that are not implemented on this agent cannot be put into a workflow"
        return arkiuse(
            binds=ReserveBindsInput(
                clients=node.binds.clients, templates=node.binds.templates
            )
            if node.binds
            else None,
            hash=node.hash,
            postman=get_current_postman(),
            provision=actor.passport.provision,
            reference=node.id,
            state_hook=actor.on_contract_change,
            assign_timeout=node.assign_timeout or None,
            yield_timeout=node.yield_timeout or None,
            reserve_timeout=node.reserve_timeout or None,
            max_retries=node.max_retries,
            retry_delay_ms=node.retry_delay,
        )  # No need to shrink inputs/outsputs for arkicontractors


async def arkimockcontractor(node: ArkitektNodeFragment, actor: Actor) -> RPCContract:
    return mockuse(
        node=node,
        provision=actor.passport.provision,
        shrink_inputs=False,
        shrink_outputs=False,
    )  # No need to shrink inputs/outputs for arkicontractors



from reaktion.actor import FlowActor
from rekuest.agents.errors import ProvisionException
from rekuest.agents.base import BaseAgent
from rekuest.actors.reactive.api import useInstanceID
import logging
from rekuest.register import register_func
from rekuest.actors.base import Actor
from rekuest.actors.types import Passport, Assignment
from rekuest.actors.transport.local_transport import (
    AgentActorTransport,
    AgentActorAssignTransport,
)
from fluss.api.schema import aget_flow
from rekuest.api.schema import aget_template, NodeKind
from rekuest.messages import Provision
from typing import Optional
from rekuest.api.schema import (
    PortInput,
    DefinitionInput,
    TemplateFragment,
    NodeKind,
    acreate_template,
    adelete_node,
    afind,
)
from fakts.fakts import Fakts
from fluss.api.schema import (
    FlowFragment,
    LocalNodeFragment,
    GraphNodeFragment,
)
from reaktion.utils import infer_kind_from_graph
from rekuest.widgets import SliderWidget, StringWidget
from rekuest.structures.default import get_default_structure_registry
from rekuest.structures.registry import StructureRegistry
from pydantic import BaseModel, Field
from .utils import convert_flow_to_definition
from rekuest.agents.extension import AgentExtension

logger = logging.getLogger(__name__)
from rekuest.definition.registry import DefinitionRegistry
from rekuest.actors.actify import reactify
from rekuest.actors.base import ActorTransport


class ReaktionExtension(BaseModel):
    structure_registry: StructureRegistry = Field(
        default_factory=get_default_structure_registry
    )

    async def aspawn_actor_from_template(
        self,
        template: TemplateFragment,
        passport: Passport,
        transport: ActorTransport,
        agent: "BaseAgent",
    ) -> Optional[Actor]:
        """Spawns an Actor from a Provision. This function closely mimics the
        spawining protocol within an actor. But maps template"""

        x = template
        assert "flow" in x.params, "Template is not a flow"

        t = await aget_flow(id=x.params["flow"])

        return FlowActor(
            flow=t,
            is_generator=x.node.kind == NodeKind.GENERATOR,
            passport=passport,
            transport=transport,
            definition=x.node,
            agent=agent,
            collector=agent.collector,
        )

    async def aregister_definitions(
        self, definition_registry: DefinitionRegistry, instance_id: str = None
    ):
        definition, actorBuilder = reactify(
            self.deploy_graph,
            self.structure_registry,
            widgets={"description": StringWidget(as_paragraph=True)},
            interfaces=["fluss:deploy"],
        )

        definition_registry.register_at_interface(
            "deploy_graph",
            definition=definition,
            structure_registry=self.structure_registry,
            actorBuilder=actorBuilder,
        )

        definition, actorBuilder = reactify(
            self.undeploy_graph,
            self.structure_registry,
            interfaces=["fluss:undeploy"],
        )

        definition_registry.register_at_interface(
            "undeploy_graph",
            definition=definition,
            structure_registry=self.structure_registry,
            actorBuilder=actorBuilder,
        )

    async def deploy_graph(
        self,
        flow: FlowFragment,
        name: str = None,
        description: str = None,
        kind: Optional[NodeKind] = None,
    ) -> TemplateFragment:
        """Deploy Flow

        Deploys a Flow as a Template

        Args:
            graph (FlowFragment): The Flow
            name (str, optional): The name of this Incarnation
            description (str, optional): The name of this Incarnation

        Returns:
            TemplateFragment: The created template
        """
        assert flow.name, "Graph must have a Name in order to be deployed"

        template = await acreate_template(
            interface=f"flow:{flow.id}",
            definition=convert_flow_to_definition(
                flow, name=name, description=description, kind=kind
            ),
            instance_id=useInstanceID(),
            params={"flow": flow.id},
            extensions=["reaktion"],
        )

        return template

    async def undeploy_graph(
        flow: FlowFragment,
    ):
        """Undeploy Flow

        Undeploys graph, no user will be able to reserve this graph anymore

        Args:
            graph (FlowFragment): The Flow

        """
        assert flow.name, "Graph must have a Name in order to be deployed"

        x = await afind(interface=flow.hash)

        await adelete_node(x)
        return None



import asyncio
from typing import List
from reaktion.atoms.transformation.base import TransformationAtom
from reaktion.events import EventType, OutEvent, InEvent
import logging
from pydantic import Field
from functools import reduce

logger = logging.getLogger(__name__)


class BufferCompleteAtom(TransformationAtom):
    buffer: List[InEvent] = Field(default_factory=list)

    async def run(self):
        try:
            while True:
                event = await self.get()

                if event.type == EventType.ERROR:
                    await self.transport.put(
                        OutEvent(
                            handle="return_0",
                            type=EventType.ERROR,
                            value=event.value,
                            source=self.node.id,
                            caused_by=[event.current_t],
                        )
                    )
                    break

                if event.type == EventType.NEXT:
                    self.buffer.append(event)

                if event.type == EventType.COMPLETE:
                    await self.transport.put(
                        OutEvent(
                            handle="return_0",
                            type=EventType.NEXT,
                            value=[
                                reduce(lambda a, b: a + list(b.value), self.buffer, [])
                            ],  # double brakcets because its  alist :)
                            source=self.node.id,
                            caused_by=[ev.current_t for ev in self.buffer],
                        )
                    )

                    await self.transport.put(
                        OutEvent(
                            handle="return_0",
                            type=EventType.COMPLETE,
                            value=[],
                            source=self.node.id,
                            caused_by=[ev.current_t for ev in self.buffer],
                        )
                    )
                    break

        except asyncio.CancelledError as e:
            logger.warning(f"Atom {self.node} is getting cancelled")
            raise e

        except Exception as e:
            logger.exception(f"Atom {self.node} excepted")
            raise e



from typing import Type


class Atom:
    pass


class Plant:
    def __init__(self) -> None:
        self.builders = {}
        pass

    def register_builder(self, atomClass: Type[Atom], type: str) -> None:
        self.builders[type] = atomClass

    def build(self, type: str, **kwargs) -> Atom:
        return self.builders[type](**kwargs)



import asyncio
from typing import Awaitable, Callable, Optional
from pydantic import BaseModel, Field
from rekuest.api.schema import AssignationLogLevel
from rekuest.messages import Assignation
from fluss.api.schema import FlowNodeCommonsFragmentBase
from reaktion.atoms.errors import AtomQueueFull
from reaktion.events import EventType, InEvent, OutEvent
import logging
from rekuest.actors.types import Assignment
from reaktion.atoms.transport import AtomTransport
from typing import Dict, Any

logger = logging.getLogger(__name__)


class Atom(BaseModel):
    node: FlowNodeCommonsFragmentBase
    transport: AtomTransport
    alog: Optional[Callable[[str, AssignationLogLevel, str], Awaitable[None]]] = Field(
        exclude=True
    )
    globals: Dict[str, Any] = Field(default_factory=dict)
    assignment: Assignment

    _private_queue: asyncio.Queue = None

    async def run(self):
        raise NotImplementedError("This needs to be implemented")

    async def get(self) -> InEvent:
        assert self._private_queue is not None, "Atom not started"
        return await self._private_queue.get()

    async def put(self, event: InEvent):
        assert self._private_queue is not None, "Atom not started"
        try:
            logger.info(f"Putting event {event}")
            await self._private_queue.put(event)  # TODO: Make put no wait?
        except asyncio.QueueFull as e:
            logger.error(f"{self.node.id} private queue is full")
            raise AtomQueueFull(f"{self.node.id} private queue is full") from e

    async def aenter(self):
        self._private_queue = asyncio.Queue()

    async def aexit(self):
        self._private_queue = None

    async def __aenter__(self):
        await self.aenter()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.aexit()

    async def start(self):
        try:
            await self.run()
        except Exception as e:
            logger.error(f"{self.node.id} FAILED", exc_info=True)
            await self.transport.put(
                OutEvent(
                    handle="return_0",
                    type=EventType.ERROR,
                    source=self.node.id,
                    value=e,
                    caused_by=[-1],
                )
            )

    @property
    def set_values(self) -> Dict[str, Any]:
        defaults = getattr(self.node, "defaults", {}) or {}
        my_globals = self.globals or {}
        return {**defaults, **my_globals}

    class Config:
        arbitrary_types_allowed = True
        underscore_attrs_are_private = True



from typing import Awaitable, Callable, Dict
from rekuest.api.schema import AssignationLogLevel, NodeKind
from rekuest.messages import Assignation
from fluss.api.schema import (
    ArkitektNodeFragment,
    LocalNodeFragment,
    FlowNodeFragment,
    ArkitektFilterNodeFragment,
    ReactiveImplementationModelInput,
    ReactiveNodeFragment,
    MapStrategy,
)
import asyncio
from reaktion.atoms.arkitekt import (
    ArkitektMapAtom,
    ArkitektMergeMapAtom,
    ArkitektAsCompletedAtom,
    ArkitektOrderedAtom,
)
from reaktion.atoms.arkitekt_filter import ArkitektFilterAtom
from reaktion.atoms.local import LocalMapAtom, LocalMergeMapAtom
from reaktion.atoms.transformation.chunk import ChunkAtom
from reaktion.atoms.transformation.buffer_complete import BufferCompleteAtom
from reaktion.atoms.transformation.split import SplitAtom
from reaktion.atoms.combination.zip import ZipAtom
from reaktion.atoms.transformation.filter import FilterAtom
from reaktion.atoms.combination.withlatest import WithLatestAtom
from reaktion.atoms.combination.gate import GateAtom
from reaktion.atoms.filter.all import AllAtom
from rekuest.postmans.utils import RPCContract
from .base import Atom
from .transport import AtomTransport
from rekuest.actors.types import Assignment
from typing import Any, Optional
from reaktion.atoms.operations.math import MathAtom, operation_map


def atomify(
    node: FlowNodeFragment,
    transport: AtomTransport,
    contract: Optional[RPCContract],
    globals: Dict[str, Any],
    assignment: Assignment,
    alog: Callable[[Assignation, AssignationLogLevel, str], Awaitable[None]] = None,
) -> Atom:
    if isinstance(node, ArkitektNodeFragment):
        if node.kind == NodeKind.FUNCTION:
            if node.map_strategy == MapStrategy.MAP:
                return ArkitektMapAtom(
                    node=node,
                    contract=contract,
                    transport=transport,
                    assignment=assignment,
                    globals=globals,
                    alog=alog,
                )
            if node.map_strategy == MapStrategy.AS_COMPLETED:
                return ArkitektAsCompletedAtom(
                    node=node,
                    contract=contract,
                    transport=transport,
                    assignment=assignment,
                    globals=globals,
                    alog=alog,
                )
            if node.map_strategy == MapStrategy.ORDERED:
                return ArkitektAsCompletedAtom(
                    node=node,
                    contract=contract,
                    transport=transport,
                    assignment=assignment,
                    globals=globals,
                    alog=alog,
                )
        if node.kind == NodeKind.GENERATOR:
            return ArkitektMergeMapAtom(
                node=node,
                contract=contract,
                transport=transport,
                assignment=assignment,
                globals=globals,
                alog=alog,
            )
    if isinstance(node, ArkitektFilterNodeFragment):
        if node.kind == NodeKind.FUNCTION:
            if node.map_strategy == MapStrategy.MAP:
                return ArkitektFilterAtom(
                    node=node,
                    contract=contract,
                    transport=transport,
                    assignment=assignment,
                    globals=globals,
                    alog=alog,
                )
        if node.kind == NodeKind.GENERATOR:
            raise NotImplementedError("Generator cannot be used as a filter")

    if isinstance(node, LocalNodeFragment):
        if node.kind == NodeKind.FUNCTION:
            return LocalMapAtom(
                node=node,
                contract=contract,
                transport=transport,
                assignment=assignment,
                globals=globals,
                alog=alog,
            )
        if node.kind == NodeKind.GENERATOR:
            return LocalMergeMapAtom(
                node=node,
                contract=contract,
                transport=transport,
                assignment=assignment,
                globals=globals,
                alog=alog,
            )

    if isinstance(node, ReactiveNodeFragment):
        if node.implementation == ReactiveImplementationModelInput.ZIP:
            return ZipAtom(
                node=node,
                transport=transport,
                assignment=assignment,
                globals=globals,
                alog=alog,
            )
        if node.implementation == ReactiveImplementationModelInput.FILTER:
            return FilterAtom(
                node=node,
                transport=transport,
                assignment=assignment,
                globals=globals,
                alog=alog,
            )
        if node.implementation == ReactiveImplementationModelInput.CHUNK:
            return ChunkAtom(
                node=node,
                transport=transport,
                assignment=assignment,
                globals=globals,
                alog=alog,
            )
        if node.implementation == ReactiveImplementationModelInput.GATE:
            return GateAtom(
                node=node,
                transport=transport,
                assignment=assignment,
                globals=globals,
                alog=alog,
            )
        if node.implementation == ReactiveImplementationModelInput.BUFFER_COMPLETE:
            return BufferCompleteAtom(
                node=node,
                transport=transport,
                assignment=assignment,
                globals=globals,
                alog=alog,
            )
        if node.implementation == ReactiveImplementationModelInput.WITHLATEST:
            return WithLatestAtom(
                node=node,
                transport=transport,
                assignment=assignment,
                globals=globals,
                alog=alog,
            )
        if node.implementation == ReactiveImplementationModelInput.COMBINELATEST:
            return WithLatestAtom(
                node=node,
                transport=transport,
                assignment=assignment,
                globals=globals,
                alog=alog,
            )
        if node.implementation == ReactiveImplementationModelInput.SPLIT:
            return SplitAtom(
                node=node,
                transport=transport,
                assignment=assignment,
                globals=globals,
                alog=alog,
            )
        if node.implementation == ReactiveImplementationModelInput.ALL:
            return AllAtom(
                node=node,
                transport=transport,
                assignment=assignment,
                globals=globals,
                alog=alog,
            )
        if node.implementation in operation_map:
            return MathAtom(
                node=node,
                transport=transport,
                assignment=assignment,
                globals=globals,
                alog=alog,
            )

    raise NotImplementedError(f"Atom for {node} is not implemented")



import asyncio
from typing import List, Tuple, Optional
from rekuest.api.schema import AssignationLogLevel
from reaktion.atoms.combination.base import CombinationAtom
from reaktion.events import EventType, OutEvent, InEvent
import logging
from pydantic import Field
from reaktion.atoms.helpers import index_for_handle
import functools

logger = logging.getLogger(__name__)


class ZipAtom(CombinationAtom):
    state: List[Optional[InEvent]] = Field(default_factory=lambda: [None, None])
    complete: List[Optional[InEvent]] = Field(default_factory=lambda: [None, None])

    async def run(self):
        self.state = list(map(lambda x: None, self.node.instream))
        self.complete = list(map(lambda x: None, self.node.instream))

        try:
            while True:
                event = await self.get()

                if event.type == EventType.ERROR:
                    await self.transport.put(
                        OutEvent(
                            handle="return_0",
                            type=EventType.ERROR,
                            value=event.value,
                            source=self.node.id,
                            caused_by=[event.current_t],
                        )
                    )
                    break

                if event.type == EventType.COMPLETE:
                    self.complete[index_for_handle(event.handle)] = event
                    if all(map(lambda x: x is not None, self.complete)):
                        await self.transport.put(
                            OutEvent(
                                handle="return_0",
                                type=EventType.COMPLETE,
                                source=self.node.id,
                                caused_by=map(lambda x: x.current_t, self.complete),
                            )
                        )
                        break

                if event.type == EventType.NEXT:
                    self.state[index_for_handle(event.handle)] = event
                    if all(map(lambda x: x is not None, self.state)):
                        await self.transport.put(
                            OutEvent(
                                handle="return_0",
                                type=EventType.NEXT,
                                source=self.node.id,
                                value=functools.reduce(
                                    lambda a, b: a + b.value, self.state, tuple()
                                ),
                                caused_by=map(lambda x: x.current_t, self.state),
                            )
                        )
                        # Reinitalizing
                        self.state = list(map(lambda x: None, self.node.instream))

        except asyncio.CancelledError as e:
            logger.warning(f"Atom {self.node} is getting cancelled")
            raise e

        except Exception:
            logger.exception(f"Atom {self.node} excepted")
